import os
import sys
import time
import random

import sqlite3
import numpy as np
import pandas as pd
import arrow as ar
from tqdm.auto import tqdm
import matplotlib.pyplot as plt


dict_growth_stages = {
    1: "Jan 1",
    2: "Jan 11",
    3: "Jan 21",
    4: "Jan 31",
    5: "Feb 10",
    6: "Feb 20",
    7: "Mar 1",
    8: "Mar 11",
    9: "Mar 21",
    10: "Mar 31",
    11: "Apr 10",
    12: "Apr 20",
    13: "Apr 30",
    14: "May 10",
    15: "May 20",
    16: "May 30",
    17: "Jun 9",
    18: "Jun 19",
    19: "Jun 29",
    20: "Jul 9",
    21: "Jul 19",
    22: "Jul 29",
    23: "Aug 8",
    24: "Aug 18",
    25: "Aug 28",
    26: "Sep 7",
    27: "Sep 17",
    28: "Sep 27",
    29: "Oct 7",
    30: "Oct 17",
    31: "Oct 27",
    32: "Nov 6",
    33: "Nov 16",
    34: "Nov 26",
    35: "Dec 6",
    36: "Dec 16",
    37: "Dec 26",
}

dict_growth_stages_biweekly = {
    1: "Jan 1",
    2: "Jan 15",
    3: "Jan 29",
    4: "Feb 12",
    5: "Feb 26",
    6: "Mar 11",
    7: "Mar 25",
    8: "Apr 8",
    9: "Apr 22",
    10: "May 6",
    11: "May 20",
    12: "Jun 3",
    13: "Jun 17",
    14: "Jul 1",
    15: "Jul 15",
    16: "Jul 29",
    17: "Aug 12",
    18: "Aug 26",
    19: "Sep 9",
    20: "Sep 23",
    21: "Oct 7",
    22: "Oct 21",
    23: "Nov 4",
    24: "Nov 18",
    25: "Dec 2",
    26: "Dec 16",
    27: "Dec 31",
}


dict_growth_stages_monthly = {
    1: "Jan 1",
    2: "Feb 1",
    3: "Mar 1",
    4: "Apr 1",
    5: "May 1",
    6: "Jun 1",
    7: "Jul 1",
    8: "Aug 1",
    9: "Sep 1",
    10: "Oct 1",
    11: "Nov 1",
    12: "Dec 1",
}

# End-date dictionaries: last day of each period (start of next period - 1 day)
dict_growth_stages_end = {
    1: "Jan 10", 2: "Jan 20", 3: "Jan 31", 4: "Feb 9", 5: "Feb 19",
    6: "Feb 28", 7: "Mar 10", 8: "Mar 20", 9: "Mar 31", 10: "Apr 9",
    11: "Apr 19", 12: "Apr 29", 13: "May 9", 14: "May 19", 15: "May 29",
    16: "Jun 8", 17: "Jun 18", 18: "Jun 28", 19: "Jul 8", 20: "Jul 18",
    21: "Jul 28", 22: "Aug 7", 23: "Aug 17", 24: "Aug 27", 25: "Sep 6",
    26: "Sep 16", 27: "Sep 26", 28: "Oct 6", 29: "Oct 16", 30: "Oct 26",
    31: "Nov 5", 32: "Nov 15", 33: "Nov 25", 34: "Dec 5", 35: "Dec 15",
    36: "Dec 25", 37: "Dec 31",
}

dict_growth_stages_biweekly_end = {
    1: "Jan 14", 2: "Jan 28", 3: "Feb 11", 4: "Feb 25", 5: "Mar 10",
    6: "Mar 24", 7: "Apr 7", 8: "Apr 21", 9: "May 5", 10: "May 19",
    11: "Jun 2", 12: "Jun 16", 13: "Jun 30", 14: "Jul 14", 15: "Jul 28",
    16: "Aug 11", 17: "Aug 25", 18: "Sep 8", 19: "Sep 22", 20: "Oct 6",
    21: "Oct 20", 22: "Nov 3", 23: "Nov 17", 24: "Dec 1", 25: "Dec 15",
    26: "Dec 30", 27: "Dec 31",
}

dict_growth_stages_monthly_end = {
    1: "Jan 31", 2: "Feb 28", 3: "Mar 31", 4: "Apr 30", 5: "May 31",
    6: "Jun 30", 7: "Jul 31", 8: "Aug 31", 9: "Sep 30", 10: "Oct 31",
    11: "Nov 30", 12: "Dec 31",
}


def statistics_file_path(dir_out, method, country, crop):
    """Build path to the per-country statistics CSV."""
    dir_statistics = dir_out / "cid" / "indices" / method / "global"
    country_str = country.title().replace("_", " ")
    crop_str = crop.title().replace("_", " ")
    return dir_statistics / f"{country_str}_{crop_str}_statistics_{method}.csv"


def remove_last_part(s):
    # Function to remove the part after the last underscore, including the last underscore
    # e.g. 'MIN_ESI4WK_33' will return 'MIN_ESI4WK'
    # 'TNx_33' will return 'TNx'
    return "_".join(s.split("_")[:-1])


def matplotlib_setup():
    import matplotlib.pyplot as plt

    plt.rcParams["font.family"] = "sans-serif"
    plt.rcParams["font.sans-serif"] = "Helvetica"

    # Set styles for axes
    plt.rcParams["axes.edgecolor"] = "#333F4B"
    plt.rcParams["axes.linewidth"] = 0.8
    plt.rcParams["xtick.color"] = "#333F4B"
    plt.rcParams["ytick.color"] = "#333F4B"


def delete_empty_dirs(_dir):
    """
    Cleanup by deleting folders which have no files in them. Delete folders which only have empty subdirs
    Args:
        _dir:

    Returns:

    """
    _dirs = [x[0] for x in os.walk(_dir)]

    for _d in _dirs:
        if not len([entry for entry in os.scandir(_d) if entry.is_file()]):
            try:
                os.removedirs(_d)
            except OSError:
                pass


def compute_zscore(val1, vals):
    """

    :param vals:
    :return:
    """
    import bottleneck as bn

    # Compute z-score for the values in the column `col`
    zscore = (val1 - bn.nanmean(vals)) / bn.nanstd(vals)

    return zscore


def add_sos(df, method, group_by):
    """

    :param df:
    :param method:
    :param group_by:
    :return:
    """
    groups = df.groupby(group_by, dropna=False)

    for key, vals in tqdm(groups, desc="Adding start of season info"):
        if not vals.empty:
            df.loc[vals.index, f"sos {method}"] = vals["Stage"].unique()[0]

    return df


def compute_zscores(
    df, method, group_by, value_column, num_years=-1, year_column="Harvest Year"
):
    """

    :param df:
    :param method:
    :param group_by:
    :param value_column:
    :param num_years: Number of years to consider for computing z-scores, -1 implies all years
    :return:
    """
    from heapq import nsmallest

    groups = df.groupby(group_by, dropna=False)

    for key, vals in tqdm(
        groups, desc=f"Computing z-scores {value_column} {method} {num_years} years"
    ):
        suffix = "" if num_years == -1 else f" {num_years} years"
        closest_years = (
            len(vals[year_column].unique()) if num_years == -1 else num_years
        )
        harvest_years = vals[year_column].unique()

        for year in [
            ar.now().year - 3,
            ar.now().year,
        ]:  # HACK only compute z-score for the last 3 years
            current_year = ar.now().year
            other_years = harvest_years[harvest_years != current_year]

            # Get the closest `num_years` years
            closest = nsmallest(closest_years, other_years, key=lambda x: abs(x - year))
            vals_subset = vals[vals[year_column].isin(closest)]
            vals_other = vals[vals[year_column] == year]

            # Compute z-score
            if not vals_subset.empty and not vals_other.empty:
                zscore = compute_zscore(
                    vals_other[value_column].values, vals_subset[value_column].values
                )
                df.loc[vals_other.index, f"Z-Score {value_column}{suffix}"] = zscore[0]

    return df


def categorize_zscores(df, bins, labels, cut_column, output_column):
    """

    :param df:
    :param bins:
    :param labels:
    :param cut_column:
    :param output_column:
    :return:
    """
    df.loc[:, output_column] = pd.cut(df[cut_column], bins=bins, labels=labels)

    return df


def detrend_column(df, column, group_by, detrended_column):
    """

    :param df:
    :param column:
    :param group_by:
    :param detrended_column:
    :return:
    """
    from scipy import signal

    groups = df.groupby(group_by, dropna=False)

    for key, vals in groups:
        # Drop rows where Yield is NaN
        vals = vals.dropna(subset=[column])

        # If removing values results in an empty dataframe, skip
        if not vals.empty:
            # Detrend Yield column and add to original dataframe
            detrended = signal.detrend(vals[column].values)
            df.loc[vals.index, detrended_column] = detrended

    return df


def categorize_column(
    df, column, group_by, zscore_column, categories, bins, category_column
):
    """
    HACk: Needs to be updated
    :param df:
    :param column:
    :param group_by:
    :param zscore_column:
    :param categories:
    :param bins:
    :param category_column:
    :return:
    """
    groups = df.groupby(group_by, dropna=False)

    for key, vals in groups:
        # Compute z-score for the values in the Detrended Yield column
        # and add to original dataframe
        zscore = compute_zscore(vals[column].values)
        df.loc[vals.index, zscore_column] = zscore

    # Categorize the z-scores
    categorize_zscores(
        df,
        bins,
        categories,
        cut_column=zscore_column,
        output_column=category_column,
    )

    return df


def get_crop_season(filename):
    """
    Get crop name and season from filename
    """
    if "winter_wheat" in filename:
        crop = "winter_wheat"
        season_index = 1
    elif "spring_wheat" in filename:
        crop = "spring_wheat"
        season_index = 1
    elif "maize" in filename:
        crop = "maize"
    elif "rice" in filename:
        crop = "rice"
    elif "soybean" in filename:
        crop = "soybean"
    elif "sorghum" in filename:
        crop = "sorghum"
    elif "millet" in filename:
        crop = "sorghum"
    elif "teff" in filename:
        crop = "teff"
    else:
        raise ValueError(f"Crop not found in {filename}")

    if "s1" in filename:
        season_index = 1
    elif "s2" in filename:
        season_index = 2
    else:
        season_index = 1

    return crop, season_index


def create_output_directory(method, admin_zone, country, crop, path_output):
    """

    :param method:
    :param admin_zone:
    :param country:
    :param crop:

    :return:
    """
    dir_output = path_output / "cid" / "indices" / method / admin_zone / country / crop
    os.makedirs(dir_output, exist_ok=True)

    return dir_output


def to_db(db_path, table_name, df, max_retries=5):
    """

    Args:
        db_path:
        table_name:
        df:
        max_retries:

    Returns:

    """
    from pangres import upsert
    from sqlalchemy import create_engine, event

    engine = create_engine(
        "sqlite:///" + str(db_path),
        connect_args={"timeout": 60},
    )

    @event.listens_for(engine, "connect")
    def set_sqlite_pragma(dbapi_connection, connection_record):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA journal_mode=WAL")
        cursor.close()

    for attempt in range(max_retries):
        try:
            upsert(
                con=engine,
                df=df,
                table_name=table_name,
                if_row_exists="update",
                chunksize=20,
            )
            return
        except Exception as e:
            if "database is locked" in str(e) and attempt < max_retries - 1:
                wait = 2 ** attempt + random.uniform(0, 1)
                print(f"DB locked writing {table_name}, retry {attempt + 1}/{max_retries} in {wait:.1f}s")
                time.sleep(wait)
            else:
                print(f"Exception: {e}")
                return


def is_table(database, table_name):
    """
    Check if table_name exists in database. Return True if it does and False if not
    Args:
        database:
        table_name:

    Returns:

    """
    con = sqlite3.connect(database)

    query = "SELECT * FROM sqlite_master"
    df = pd.read_sql_query(query, con)
    con.close()

    if table_name in df["tbl_name"].values:
        return True
    else:
        return False


def plot(X, labels, probabilities=None, parameters=None, ground_truth=False, ax=None):
    if ax is None:
        _, ax = plt.subplots(figsize=(10, 4))
    labels = labels if labels is not None else np.ones(X.shape[0])
    probabilities = probabilities if probabilities is not None else np.ones(X.shape[0])
    # Black removed and is used for noise instead.
    unique_labels = set(labels)
    colors = [plt.cm.Spectral(each) for each in np.linspace(0, 1, len(unique_labels))]
    # The probability of a point belonging to its labeled cluster determines
    # the size of its marker
    proba_map = {idx: probabilities[idx] for idx in range(len(labels))}
    for k, col in zip(unique_labels, colors):
        if k == -1:
            # Black used for noise.
            col = [0, 0, 0, 1]

        class_index = np.where(labels == k)[0]
        for ci in class_index:
            try:
                ax.plot(
                    X[ci, 0],
                    X[ci, 1],
                    "x" if k == -1 else "o",
                    markerfacecolor=tuple(col),
                    markeredgecolor="k",
                    markersize=4 if k == -1 else 1 + 5 * proba_map[ci],
                )
            except:
                breakpoint()
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    preamble = "True" if ground_truth else "Estimated"
    title = f"{preamble} number of clusters: {n_clusters_}"
    if parameters is not None:
        parameters_str = ", ".join(f"{k}={v}" for k, v in parameters.items())
        title += f" | {parameters_str}"
    ax.set_title(title)
    plt.tight_layout()
    plt.show()


def pairwise_rmse(df1, df2):
    return np.sqrt(((df1 - df2) ** 2).mean())


def nse(observed, simulated, weights=None):
    """
    Compute Weighted Nash-Sutcliffe Efficiency
    :param observed: Array of observed values
    :param simulated: Array of simulated values
    :param weights: Optional array of weights
    :return: NSE value
    """
    if weights is None:
        weights = np.ones_like(observed)

    weighted_sq_diff = np.sum(weights * (observed - simulated) ** 2)
    weighted_variance = np.sum(
        weights * (observed - np.average(observed, weights=weights)) ** 2
    )

    return 1 - (weighted_sq_diff / weighted_variance)


def mape(observed, simulated, weights=None):
    """
    Compute Weighted Mean Absolute Percentage Error
    :param observed: Array of observed values
    :param forecast: Array of forecast values
    :param weights: Optional array of weights
    :return: MAPE value
    """
    if weights is None:
        weights = np.ones_like(observed)

    weighted_abs_percent_error = weights * np.abs((observed - simulated) / observed)

    return np.sum(weighted_abs_percent_error) / np.sum(weights) * 100


def pbias(observed, simulated, weights=None):
    """
    Compute Weighted Percent Bias
    :param observed: Array of observed values
    :param simulated: Array of simulated values
    :param weights: Optional array of weights
    :return: PBIAS value
    """
    if weights is None:
        weights = np.ones_like(observed)

    weighted_diff_sum = np.sum(weights * (observed - simulated))
    weighted_obs_sum = np.sum(weights * observed)

    return (weighted_diff_sum / weighted_obs_sum) * 100


# Function to remove trend
def detrend(data):
    return data.diff().dropna()


# Function to add trend back
def retrend(original_data, detrended_data):
    try:
        retrended_data = detrended_data.cumsum() + original_data.iloc[0]
    except:
        breakpoint()

    return retrended_data


def linregress(x, y):
    """ """
    # Fit a linear regression model using statsmodels
    import statsmodels.api as sm

    x = sm.add_constant(x)
    model = sm.OLS(y, x).fit()

    # Get the slope, intercept, p-value of the trendline
    intercept = model.params[0]
    slope = model.params[1]
    p = model.f_pvalue

    if p < 0.05:
        # Significant, therefore add * to the equation
        eqn = f"y = {slope:.3f}x + {intercept:.2f} *"
    else:
        eqn = f"y = {slope:.3f}x + {intercept:.2f}"

    return eqn


def slope(x, y):
    """ """
    from scipy.stats import mstats
    import pymannkendall as mk

    # Reset index to ensure x and y are aligned
    x = x.reset_index(drop=True)
    y = y.reset_index(drop=True)

    # Note that we are getting slope from one library and intercept from another
    # This is because pymannkendall reports intercept of the Kendall-Theil Robust Line
    # and theilslopes reports intercept of the Theil-Sen estimator, we want the latter
    try:
        trend, h, p, z, Tau, s, var_s, slope, intercept = mk.original_test(y)
        slope, intercept = mstats.theilslopes(y, x)[0], mstats.theilslopes(y, x)[1]
    except (ValueError, ZeroDivisionError):
        slope = np.nan
        intercept = np.nan
        p = np.nan

    return slope, intercept, p


def is_trending(x, y, threshold=0.05):
    """ """
    import pymannkendall as mk

    # Reset index to ensure x and y are aligned
    x = x.reset_index(drop=True)
    y = y.reset_index(drop=True)

    trend, h, p, z, Tau, s, var_s, slope, intercept = mk.original_test(y)

    if p < threshold:
        return True
    else:
        return False


def process_subsets(df, custom_function, **kwargs):
    """
    Processes subsets of the dataframe based on unique values in two columns and
    applies a given custom function

    :param df (pd.DataFrame): The dataframe to process
    :param custom_function: A function to apply to each subsubset dataframe
    :param **kwargs  Additional keyword arguments to pass to the operation function
    """
    frames = []
    for i, value1 in enumerate(df[kwargs["column1"]].unique()):
        subset_df = df[df[kwargs["column1"]] == value1]

        if not subset_df.empty:
            if kwargs["column2"]:
                for j, value2 in enumerate(subset_df[kwargs["column2"]].unique()):
                    subsubset_df = subset_df[subset_df[kwargs["column2"]] == value2]

                    results = custom_function(subsubset_df, i, j, **kwargs)
                    frames.append(results)
            else:
                results = custom_function(subsubset_df, i, i, **kwargs)
                frames.append(results)

    df_results = pd.concat(frames)

    return df_results


def compute_dataframe_transformation(df, group_by=[], column=None, stat="mean"):
    """

    :param df:
    :param group_by:
    :param column:
    :param stat:
    """
    if stat == "mean":
        df = df.groupby(group_by, dropna=False)[column].mean().reset_index()
    else:
        raise NotImplementedError(f"stat {stat} not implemented")

    # Drop any rows where column is NaN
    df = df.dropna(subset=[column])

    return df


def list_directories(directory_path):
    """
    Lists all directories within the specified directory path.

    Parameters:
    - directory_path: A string representing the path to the directory you want to explore.

    Returns:
    - A list of directory names found within the specified directory.
    """
    try:
        # List all entries in the directory given by "directory_path"
        directory_contents = os.listdir(directory_path)

        # Filter out the directories from all entries
        directories = [
            d
            for d in directory_contents
            if os.path.isdir(os.path.join(directory_path, d))
        ]

        return directories
    except FileNotFoundError:
        print(f"Error: The directory '{directory_path}' was not found.")
        return []
    except PermissionError:
        print(f"Error: Permission denied to access the directory '{directory_path}'.")
        return []
    except Exception as e:
        print(f"An error occurred: {e}")
        return []


def compute_biweekly_index(row):
    """
    Compute the index of the biweekly period for a given day of the year.

    The first biweekly period of the year starts on January 1st.
    Args:
    - row (pd.Series): A row from the DataFrame containing 'year' and 'doy' columns.

    Returns:
    - int: The biweekly period index (starting from 1).
    """
    # Calculate the day of the year, adjusting to start from 0
    day_of_year_zero_indexed = int(row["Doy"]) - 1
    # Compute the biweekly index, adjusting so the first period is index 1
    biweekly_index = (day_of_year_zero_indexed // 14) + 1

    return biweekly_index


def is_leap_year(year):
    """Check if the specified year is a leap year."""
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def compute_time_periods(df_inp, period_type, year):
    """
    Compute the starting and ending time-periods for a given period type ('dekad', 'biweekly', 'monthly')
    in either a leap or non-leap year.

    Parameters:
    - df: DataFrame with a 'Doy' column.
    - period_type: String indicating the period type ('dekad', 'biweekly', 'monthly').
    - year: The year for the Doy values, used to handle leap years.

    Returns:
    - A tuple with starting and ending periods for the specified period type.
    """
    # Adjust the origin based on whether the year is a leap year
    df = df_inp.copy()
    df["Date"] = pd.to_datetime(
        df["Doy"], unit="D", origin=pd.Timestamp(f"{year}-01-01")
    )

    if period_type.startswith("dekad"):
        df["Period"] = ((df["Date"].dt.dayofyear - 1) // 10) + 1
    elif period_type.startswith("biweekly"):
        df["Period"] = ((df["Date"].dt.dayofyear - 1) // 14) + 1
    elif period_type.startswith("monthly"):
        df["Period"] = df["Month"]
    else:
        return "Invalid period type. Choose 'dekad', 'biweekly', or 'monthly'."

    start_period = df.iloc[0]["Period"]
    end_period = df.iloc[-1]["Period"]

    return start_period, end_period


def compute_h_index(values):
    # Sort the array in descending order
    sorted_value = np.sort(values)[::-1]

    # Iterate through the sorted array to find the h-index
    h_index = 0

    for i, value in enumerate(sorted_value, start=1):
        if value >= i:
            h_index = value
        else:
            break

    return h_index


def get_z_value(alpha):
    """
    Calculate the z-value for a given alpha level.

    Parameters:
    alpha (float): The significance level (e.g., 0.05 for a 95% confidence interval)

    Returns:
    float: The corresponding z-value
    """
    from scipy.stats import norm

    return norm.ppf(1 - alpha / 2)


def wait_or_keypress(seconds=20):
    """Wait for *seconds* or until the user presses any key (cross-platform)."""
    try:
        if sys.platform == "win32":
            import msvcrt
            for remaining in range(seconds, 0, -1):
                print(f"\r  Starting in {remaining}s ... (press any key to start now) ", end="", flush=True)
                deadline = time.monotonic() + 1.0
                while time.monotonic() < deadline:
                    if msvcrt.kbhit():
                        msvcrt.getch()
                        print("\r" + " " * 60 + "\r", end="", flush=True)
                        return
                    time.sleep(0.05)
        else:
            import select
            for remaining in range(seconds, 0, -1):
                print(f"\r  Starting in {remaining}s ... (press Enter to start now) ", end="", flush=True)
                rlist, _, _ = select.select([sys.stdin], [], [], 1.0)
                if rlist:
                    sys.stdin.readline()
                    print("\r" + " " * 60 + "\r", end="", flush=True)
                    return
    except Exception:
        time.sleep(seconds)

    print("\r" + " " * 60 + "\r", end="", flush=True)


def display_run_summary(title, params, wait=20):
    """
    Print a rich-formatted summary of run parameters, then wait.

    Args:
        title: Header string (e.g. "GeoCIF ML Runner" or "CEI Indices Runner")
        params: list of (label, value) tuples to display
        wait: seconds to wait before starting (0 to skip)
    """
    from rich.console import Console
    from rich.panel import Panel
    from rich.table import Table
    from geocif import __version__

    console = Console()
    table = Table(show_header=False, box=None, padding=(0, 1))
    table.add_column(style="bold cyan", no_wrap=True)
    table.add_column()

    table.add_row("version", __version__)
    for label, value in params:
        if isinstance(value, list):
            value = ", ".join(str(v) for v in value)
        table.add_row(label, str(value))

    console.print()
    console.print(
        Panel(
            table,
            title=f"[bold bright_white]{title}[/bold bright_white]",
            border_style="bright_blue",
            padding=(1, 2),
        )
    )
    console.print()

    if wait > 0:
        wait_or_keypress(wait)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# 1. Read the CSV
df = pd.read_csv(r"C:\Users\ritvik\Downloads\ET_AgStats.csv")

# 2. Filter for the crop "Maize (Corn)"
df = df[df['DNL_SourceCrop'] == 'Maize (Corn)']

# 3. Remove rows where "Area Planted: ha" is "NA" or "NC"
df = df[df['Area Planted: ha'] != 'NA']
df = df[df['Area Planted: ha'] != 'NC']

df = df[df['Yield: MT/ha'] != 'NA']
df = df[df['Yield: MT/ha'] != 'NC']
df = df[df['Yield: MT/ha'] != '#REF!']

# Remove rows where Admin 2 is null
df = df.dropna(subset=['Admin 2'])
df = df.dropna(subset=['Yield: MT/ha'])

# 4. Convert "Area Planted: ha" to float by removing commas
df['Area Planted: ha'] = (
    df['Area Planted: ha']
    .str.replace(',', '', regex=False)
    .astype(float)
)

df['Yield: MT/ha'] = (
    df['Yield: MT/ha']
    .str.replace(',', '', regex=False)
    .astype(float)
)

# 5. Group by [region, season] to calculate z-scores
grouped = df.groupby(['Admin 2', 'Season'])
anomalies_list = []

for (region, season), group_data in grouped:
    mean_area = group_data['Area Planted: ha'].mean()
    std_area = group_data['Area Planted: ha'].std()

    # Avoid division by zero
    if std_area == 0:
        group_data['Z_score'] = 0
    else:
        group_data['Z_score'] = (group_data['Area Planted: ha'] - mean_area) / std_area

    # Flag anomalies if abs(z-score) > 3
    group_data['Anomaly'] = group_data['Z_score'].apply(lambda x: 'Yes' if abs(x) > 3 else 'No')

    anomalies_list.append(group_data)

# 6. Concatenate grouped data back together
df_analyzed = pd.concat(anomalies_list)

# 7. Filter to see only anomalies
df_anomalies = df_analyzed[df_analyzed['Anomaly'] == 'Yes']

# 8. Print full dataset with anomaly flags and the subset of anomalies
print("All data with anomaly flags:")
print(df_analyzed)

print("\nDetected anomalies:")
print(df_anomalies)
df_anomalies.to_csv(r"df_anomalies_v2.csv", index=False)

# 11. Distribution of "Yield: MT/ha"

plt.figure(figsize=(8, 5))
sns.histplot(df['Yield: MT/ha'], kde=True, bins=30)
plt.title('Distribution of Yield (MT/ha)')
plt.xlabel('Yield (MT/ha)')
plt.ylabel('Count')
plt.tight_layout()
plt.show()

# count number of values where yield < 1
low_yield = df[df['Yield: MT/ha'] < 1].shape[0]
total = df.shape[0]
print(f"Number of records with yield < 1: {low_yield} / {total}")
breakpoint()
# 9. Bar chart of number of anomalies per Season
anomalies_by_season = df_anomalies['Season'].value_counts()
plt.figure(figsize=(8, 5))
anomalies_by_season.plot(kind='bar')
plt.title('Number of Anomalies per Season')
plt.xlabel('Season')
plt.ylabel('Count of Anomalies')
plt.tight_layout()
plt.show()

# 10. Heatmap of anomalies by Region (rows) and Year (columns)

# Ensure "Year" is numeric for pivoting
df_anomalies['Year'] = pd.to_numeric(df_anomalies['Year'], errors='coerce')

# Count how many anomalies per (region, year)
heatmap_data = df_anomalies.groupby(['Admin 1', 'Year']).size().unstack(fill_value=0)

# Plot the heatmap
plt.figure(figsize=(10, 6))
sns.heatmap(
    heatmap_data,
    annot=True,
    cmap='Blues',
    fmt='d'
)
plt.title('Number of Anomalies by Region and Year')
plt.xlabel('Year')
plt.ylabel('Region')
plt.tight_layout()
plt.show()





import os
import ast
import multiprocessing as mp
from pathlib import Path

from tqdm import tqdm
import matplotlib.pyplot as plt

from geocif import logger as log
from geocif import utils as ut
from .ml import output
from geocif import geocif

plt.style.use("default")

# Show usage info on import
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

_console = Console()
_table = Table(show_header=False, box=None, padding=(0, 1))
_table.add_column(style="bold cyan", no_wrap=True)
_table.add_column()
_table.add_row("Usage", "from geocif import geocif_runner; geocif_runner.run(cfg)")
_table.add_row("cfg", "\\[geobase.txt, countries.txt, crops.txt, geocif.txt]")
_console.print(Panel(_table, title="[bold bright_white]GeoCIF ML Runner[/]", border_style="bright_blue", padding=(1, 2)))


def _loop_execute(logger, parser, project_name, country, crop, season, model, index):
    """Execute ML pipeline for a single country/crop/season/model combination."""
    obj = geocif.Geocif(logger=logger, parser=parser, project_name=project_name)
    obj.read_data(country, crop, season)

    if not hasattr(obj, 'df_inputs') or obj.df_inputs is None:
        return

    # Store config file in database, only execute this for
    # the first iteration of the loop
    if index == 0:
        output.config_to_db(obj.db_path, obj.parser, obj.today)

    # Setup metadata and run ML code
    obj.setup(season, model)
    if obj.simulation_stages:
        obj.execute()


def _loop_execute_pooled(logger, parser, project_name, countries, crop, season, model, index):
    """Execute ML pipeline with pooled data from multiple countries."""
    obj = geocif.Geocif(logger=logger, parser=parser, project_name=project_name)
    obj.read_data_pooled(countries, crop, season)

    if not hasattr(obj, 'df_inputs') or obj.df_inputs is None:
        return

    if index == 0:
        output.config_to_db(obj.db_path, obj.parser, obj.today)

    obj.setup_pooled(countries, season, model)
    if obj.simulation_stages:
        obj.execute()


def loop_execute(inputs):
    """Unpack inputs and run single-country ML pipeline."""
    project_name, country, crop, season, model, logger, parser, index = inputs

    logger.info("=====================================================")
    logger.info(f"\tStarting GEOCIF: {country} {crop} {season} {model}")
    logger.info("=====================================================")

    _loop_execute(logger, parser, project_name, country, crop, season, model, index)


def loop_execute_pooled(inputs):
    """Unpack inputs and run pooled multi-country ML pipeline."""
    project_name, countries, crop, season, model, logger, parser, index = inputs

    logger.info("=====================================================")
    logger.info(f"\tStarting GEOCIF (pooled): {countries} {crop} {season} {model}")
    logger.info("=====================================================")

    _loop_execute_pooled(logger, parser, project_name, countries, crop, season, model, index)


def gather_inputs(parser):
    """Build list of [project_name, country, crop, season, model] tuples."""
    project_name = parser.get("DEFAULT", "project_name")
    countries = ast.literal_eval(parser.get("DEFAULT", "countries"))

    all_inputs = []
    for country in countries:
        for crop in ast.literal_eval(parser.get(country, "crops")):
            for season in ast.literal_eval(parser.get(country, "forecast_seasons")):
                for model in ast.literal_eval(parser.get(country, "models")):
                    all_inputs.append([project_name, country, crop, season, model])

    return all_inputs


def gather_pooled_inputs(parser):
    """Group inputs by (crop, season, model) for cross-country pooling.

    Returns list of [project_name, [country1, ...], crop, season, model].
    """
    project_name = parser.get("DEFAULT", "project_name")
    countries = ast.literal_eval(parser.get("DEFAULT", "countries"))

    groups = {}
    for country in countries:
        for crop in ast.literal_eval(parser.get(country, "crops")):
            for season in ast.literal_eval(parser.get(country, "forecast_seasons")):
                for model in ast.literal_eval(parser.get(country, "models")):
                    groups.setdefault((crop, season, model), []).append(country)

    return [[project_name, clist, crop, season, model]
            for (crop, season, model), clist in groups.items()]


def execute_models(inputs, logger, parser, loop_fn=None):
    """
    Executes the model either in parallel or serially based on configuration.

    Args:
        inputs (list): The input data for model execution.
        logger (logging.Logger): Logger for tracking execution details
        parser (configparser.ConfigParser): Configuration file parser
        loop_fn (callable): Function to call per input. Defaults to loop_execute.
    """
    if loop_fn is None:
        loop_fn = loop_execute

    do_parallel = parser.getboolean("DEFAULT", "do_parallel_ml", fallback=False)

    # Add logger and parser to each element in inputs
    inputs = [item + [logger, parser, idx] for idx, item in enumerate(inputs)]

    if do_parallel:
        fraction_cpus = parser.getfloat("DEFAULT", "fraction_cpus")
        cpu_count = int(mp.cpu_count() * fraction_cpus)

        with mp.Pool(cpu_count) as pool:
            pool.map(loop_fn, inputs)
    else:
        pbar = tqdm(inputs, desc="Executing ML models")
        for item in pbar:
            crop, season = item[2], item[3]
            label = item[1] if isinstance(item[1], str) else "pooled"
            pbar.set_description(f"{label} {crop} {season}")
            loop_fn(item)

    logger.info("======================================")
    logger.info("\tCompleted all model executions")
    logger.info("======================================")


def _build_summary_params(parser, inputs):
    """Build parameter list for the run summary display."""
    countries = ast.literal_eval(parser.get("DEFAULT", "countries"))
    if parser.has_option("DEFAULT", "do_parallel_ml"):
        do_parallel = parser.getboolean("DEFAULT", "do_parallel_ml")
    else:
        do_parallel = False

    # Collect unique crops, seasons, models per country
    country_details = {}
    for item in inputs:
        _, country_or_list, crop, season, model = item[:5]
        # In pooled mode, country_or_list is a list of countries
        if isinstance(country_or_list, list):
            for c in country_or_list:
                info = country_details.setdefault(c, {"crops": set(), "seasons": set(), "models": set()})
                info["crops"].add(crop)
                info["seasons"].add(str(season))
                info["models"].add(model)
        else:
            info = country_details.setdefault(country_or_list, {"crops": set(), "seasons": set(), "models": set()})
            info["crops"].add(crop)
            info["seasons"].add(str(season))
            info["models"].add(model)

    params = [("Countries", countries)]
    for country, info in country_details.items():
        params.append((f"  {country} crops", sorted(info["crops"])))
        params.append((f"  {country} seasons", sorted(info["seasons"])))
        params.append((f"  {country} models", sorted(info["models"])))

    # Global settings
    for key in ["db", "method"]:
        if parser.has_option("DEFAULT", key):
            params.append((key, parser.get("DEFAULT", key)))

    # ML settings (safe reads with fallbacks)
    for key, section in [
        ("feature_selection", "ML"),
        ("cluster_strategy", "ML"),
        ("model_type", "ML"),
        ("check_yield_trend", "ML"),
        ("use_single_time_period_as_feature", "ML"),
        ("lag_yield_as_feature", "ML"),
        ("use_spatial_neighbors", "ML"),
    ]:
        if parser.has_option(section, key):
            params.append((key, parser.get(section, key)))

    # Per-model use_ceis (show from first model)
    first_model = inputs[0][4] if inputs else None  # index 4 = model name
    if first_model and parser.has_option(first_model, "use_ceis"):
        params.append(("use_ceis", parser.get(first_model, "use_ceis")))

    params.append(("Parallel", str(do_parallel)))
    if do_parallel:
        fraction_cpus = parser.getfloat("DEFAULT", "fraction_cpus")
        cpu_count = int(mp.cpu_count() * fraction_cpus)
        params.append(("CPUs", str(cpu_count)))
    params.append(("Total combinations", str(len(inputs))))

    return params


def main(logger, parser):
    """Run the GeoCIF ML pipeline."""
    from geocif.data import ensure_metadata
    ensure_metadata(parser)

    pool_countries = parser.getboolean("ML", "pool_countries", fallback=False)

    if pool_countries:
        inputs = gather_pooled_inputs(parser)
        params = _build_summary_params(parser, inputs)
        ut.display_run_summary("GeoCIF ML Runner (Pooled)", params, wait=20)
        execute_models(inputs, logger, parser, loop_fn=loop_execute_pooled)
    else:
        inputs = gather_inputs(parser)
        params = _build_summary_params(parser, inputs)
        ut.display_run_summary("GeoCIF ML Runner", params, wait=20)
        execute_models(inputs, logger, parser)


def run(path_config_files=[Path("../config/geocif.txt")]):
    logger, parser = log.setup_logger_parser(path_config_files)
    main(logger, parser)


if __name__ == "__main__":
    run()



"""Yield Outlook Map Generator.

Runs the ML pipeline for the current year and last N historical years,
then generates a diverging choropleth map showing current forecast yield
as a percentage of the historical mean/median prediction per region.
"""

import ast
import logging
import os
import sqlite3
import warnings
from pathlib import Path

import arrow as ar
import geopandas as gpd
import numpy as np
import palettable as pal
import pandas as pd

from geocif import geocif_runner as gc
from geocif import logger as log
from geocif import utils as ut
from .viz import plot

warnings.simplefilter(action="ignore", category=FutureWarning)

logger = logging.getLogger(__name__)

# Show usage info on import
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

_console = Console()
_table = Table(show_header=False, box=None, padding=(0, 1))
_table.add_column(style="bold cyan", no_wrap=True)
_table.add_column()
_table.add_row("Usage", "from geocif import yield_outlook; yield_outlook.run(cfg)")
_table.add_row("cfg", "\\[geobase.txt, countries.txt, crops.txt, geocif.txt]")
_table.add_row("reuse_db", "yield_outlook.run(cfg, reuse_db='/path/to/outlook_MM_DD_YYYY.db')")
_console.print(
    Panel(
        _table,
        title="[bold bright_white]GeoCIF Yield Outlook[/]",
        border_style="bright_blue",
        padding=(1, 2),
    )
)


def _load_shapefiles(parser):
    """Load and concatenate shapefiles for all countries.

    Reuses the shapefile loading pattern from analysis.py Geoanalysis.setup().

    Returns:
        dg: Combined GeoDataFrame with 'Country Region' merge column.
        dict_config: Per country_crop config dict with method, crops, models, etc.
    """
    from geoprepare.georegion import get_boundary_col_mapping

    countries = ast.literal_eval(parser.get("DEFAULT", "countries"))
    dir_boundary_files = Path(parser.get("PATHS", "dir_boundary_files"))
    pool_countries = parser.getboolean("ML", "pool_countries", fallback=False)

    dict_config = {}
    all_shapefiles = []

    for country in countries:
        crops = ast.literal_eval(parser.get(country, "crops"))
        models = ast.literal_eval(parser.get(country, "models"))
        method = parser.get(country, "method")
        admin_zone = parser.get(country, "admin_level")

        for crop in crops:
            dict_config[f"{country}_{crop}"] = {
                "method": method,
                "crops": crop,
                "models": models,
                "admin_zone": admin_zone,
            }

        # Load shapefile
        shp_file = parser.get(country, "boundary_file")
        dg_country = gpd.read_file(
            dir_boundary_files / shp_file,
            engine="pyogrio",
        )

        # Rename columns using config-driven mapping
        rename = get_boundary_col_mapping(parser, shp_file)
        targets = set(rename.values())
        sources = set(rename.keys())
        conflicting = [
            c for c in dg_country.columns if c in targets and c not in sources
        ]
        if conflicting:
            dg_country = dg_country.drop(columns=conflicting)
        dg_country = dg_country.rename(columns=rename)

        if "ADM0_NAME" not in dg_country.columns:
            dg_country.loc[:, "ADM0_NAME"] = country.title().replace("_", " ")

        all_shapefiles.append(dg_country)

    # Add pooled table entries when pool_countries is enabled
    if pool_countries:
        all_crops = set()
        for country in countries:
            crops = ast.literal_eval(parser.get(country, "crops"))
            all_crops.update(crops)
        first_models = ast.literal_eval(parser.get(countries[0], "models"))
        first_method = parser.get(countries[0], "method")
        first_admin = parser.get(countries[0], "admin_level")
        for crop in all_crops:
            dict_config[f"pooled_{crop}"] = {
                "method": first_method,
                "crops": crop,
                "models": first_models,
                "admin_zone": first_admin,
            }

    dg = pd.concat(all_shapefiles, ignore_index=True)

    # Create "Country Region" merge column (same pattern as analysis.py:1229-1242)
    dg["Country Region"] = dg["ADM0_NAME"]
    dg["Country Region"] = dg["Country Region"].str.cat(dg["ADM1_NAME"], sep=" ")
    if "ADM2_NAME" in dg.columns:
        dg.loc[dg["ADM2_NAME"].notna(), "Country Region"] = (
            dg["ADM0_NAME"] + " " + dg["ADM2_NAME"]
        )
    dg["Country Region"] = dg["Country Region"].str.lower().replace("_", " ")

    return dg, dict_config


def _query_predictions(db_path, table, model, experiment_name="default"):
    """Query predictions from the SQLite database for a specific model.

    Returns DataFrame with columns: Country, Region, Harvest Year, Stage Name,
    Predicted Yield (tn per ha).
    """
    if not db_path.exists():
        logger.error("Database not found: %s", db_path)
        return pd.DataFrame()

    con = sqlite3.connect(db_path)
    try:
        df = pd.read_sql(
            f'SELECT "Country", "Region", "Harvest Year", "Stage Name", '
            f'"Predicted Yield (tn per ha)" '
            f'FROM "{table}" WHERE "Experiment Name" = ? AND "Model" = ?',
            con,
            params=(experiment_name, model),
        )
    except (pd.errors.DatabaseError, sqlite3.OperationalError) as e:
        logger.warning("Failed to query table '%s': %s", table, e)
        df = pd.DataFrame()
    finally:
        con.close()
    if not df.empty:
        if "Harvest Year" in df.columns:
            df["Harvest Year"] = df["Harvest Year"].astype(int)
        if "Predicted Yield (tn per ha)" in df.columns:
            df["Predicted Yield (tn per ha)"] = pd.to_numeric(
                df["Predicted Yield (tn per ha)"], errors="coerce"
            )
    return df


def _compute_outlook_index(df, current_year, n_years, aggregation,
                           use_latest_stage=True, stage_name=None):
    """Compute yield outlook index per region.

    Args:
        use_latest_stage: If True (default), use the latest stage per
            (Country, Region, Harvest Year) so stage name mismatches
            between current and historical years don't matter.
            If False, filter by exact stage_name.

    outlook_index = (current_predicted / agg(historical_predicted)) * 100

    Returns DataFrame with columns: Country, Region, Country Region,
    current_predicted, hist_predicted, outlook_index.
    """
    if use_latest_stage:
        # For each region+year, keep only the latest (last) stage prediction
        df_work = (
            df.sort_values("Stage Name")
            .groupby(["Country", "Region", "Harvest Year"])
            .last()
            .reset_index()
        )
    else:
        df_work = df[df["Stage Name"] == stage_name].copy()

    # Current year predictions per region
    df_current = df_work[df_work["Harvest Year"] == current_year]
    current_pred = (
        df_current.groupby(["Country", "Region"])["Predicted Yield (tn per ha)"]
        .mean()
        .rename("current_predicted")
    )

    # Historical years per region
    min_year = current_year - n_years
    df_hist = df_work[
        (df_work["Harvest Year"] < current_year)
        & (df_work["Harvest Year"] >= min_year)
    ]
    agg_func = "median" if aggregation == "median" else "mean"
    hist_agg = (
        df_hist.groupby(["Country", "Region"])["Predicted Yield (tn per ha)"]
        .agg(agg_func)
        .rename("hist_predicted")
    )

    # Compute index
    df_outlook = pd.concat([current_pred, hist_agg], axis=1).dropna()
    df_outlook["outlook_index"] = np.where(
        df_outlook["hist_predicted"] != 0,
        (df_outlook["current_predicted"] - df_outlook["hist_predicted"])
        / df_outlook["hist_predicted"] * 100,
        np.nan,
    )
    df_outlook = df_outlook.reset_index()

    # Create merge column (same pattern as analysis.py:1410-1414)
    df_outlook["Country Region"] = (
        df_outlook["Country"].str.lower().str.replace("_", " ")
        + " "
        + df_outlook["Region"].str.lower()
    )

    return df_outlook


def _generate_outlook_map(
    dg,
    df_outlook,
    countries,
    crop,
    model,
    current_year,
    n_years,
    aggregation,
    dir_out,
    stage_name="",
    annotate_regions=False,
):
    """Generate a diverging choropleth map of the yield outlook index."""
    col = "outlook_index"

    # Fixed range: -40% to +40% departure (matching analysis.py anomaly maps)
    vmin = -40
    vmax = 40

    # Determine extend arrows based on actual data range
    data_min = df_outlook[col].min()
    data_max = df_outlook[col].max()
    if data_min < vmin and data_max > vmax:
        extend = "both"
    elif data_min < vmin:
        extend = "min"
    elif data_max > vmax:
        extend = "max"
    else:
        extend = "neither"

    countries_display = [c.title().replace("_", " ") for c in countries]
    stage_suffix = f"_{stage_name}" if stage_name else ""
    if len(countries) > 1:
        fname = f"yield_outlook_{len(countries)}_countries_{crop}_{model}{stage_suffix}_{current_year}.png"
    else:
        fname = f"yield_outlook_{'_'.join(countries)}_{crop}_{model}{stage_suffix}_{current_year}.png"

    stage_label = f", stage {stage_name}" if stage_name else ""
    plot.plot_map(
        dg,
        df_outlook,
        merge_col="Country Region",
        name_country=countries_display,
        name_col=col,
        dir_out=dir_out,
        fname=fname,
        label=f"% departure from {n_years}-yr {aggregation}\n{crop.title()}, {current_year}{stage_label}",
        vmin=vmin,
        vmax=vmax,
        cmap=pal.colorbrewer.diverging.BrBG_11,
        series="diverging",
        annotate_regions=annotate_regions,
        loc_legend="lower left",
        extend=extend,
    )


def run(path_config_files=None, current_year=None, n_years=None, aggregation=None,
        reuse_db=None, use_latest_stage=True):
    """Main entry point for yield outlook map generation.

    1. Override forecast_seasons to cover [current_year - n_years, ..., current_year]
    2. Run the ML pipeline via gc.execute_models()
    3. Query predictions from the database
    4. Compute outlook index per region
    5. Generate diverging choropleth maps and CSV

    Args:
        path_config_files: List of config file paths.
        current_year: Forecast year (default: this year).
        n_years: Number of historical years for comparison (default: config or 10).
        aggregation: 'mean' or 'median' (default: config or 'mean').
        reuse_db: Path to existing outlook DB to skip ML and regenerate maps only.
        use_latest_stage: If True (default), use latest available stage per
            region+year. Handles mismatched stage names across years/countries.
    """
    if path_config_files is None:
        path_config_files = [Path("../config/geocif.txt")]

    logger_obj, parser = log.setup_logger_parser(path_config_files)

    # Read config with defaults
    if n_years is None:
        n_years = parser.getint("ML", "outlook_n_years", fallback=10)
    if aggregation is None:
        aggregation = parser.get("ML", "outlook_aggregation", fallback="mean")
    if current_year is None:
        current_year = ar.utcnow().to("America/New_York").year

    countries = ast.literal_eval(parser.get("DEFAULT", "countries"))
    experiment_name = parser.get("DEFAULT", "experiment_name", fallback="default")

    if reuse_db is not None:
        # ---- Skip ML, reuse existing DB ----
        reuse_path = Path(reuse_db)
        if not reuse_path.exists():
            logger.error("Reuse DB not found: %s", reuse_path)
            return
        outlook_db = reuse_path.name
        logger.info("Reusing existing outlook DB: %s", reuse_path)
    else:
        # ---- Step 1: Run ML pipeline for all needed years ----
        outlook_seasons = list(range(current_year - n_years, current_year + 1))
        originals = {}
        for country in countries:
            originals[country] = parser.get(country, "forecast_seasons")
            parser.set(country, "forecast_seasons", str(outlook_seasons))

        parser.set("DEFAULT", "experiment_name", "outlook")
        orig_db = parser.get("DEFAULT", "db")
        outlook_db = ar.utcnow().to("America/New_York").format("[outlook_]MM[_]DD[_]YYYY[.db]")
        parser.set("DEFAULT", "db", outlook_db)
        orig_parallel = parser.get("DEFAULT", "do_parallel_ml", fallback=None)
        parser.set("DEFAULT", "do_parallel_ml", "False")

        pool_countries_flag = parser.getboolean("ML", "pool_countries", fallback=False)
        if pool_countries_flag:
            inputs = gc.gather_pooled_inputs(parser)
        else:
            inputs = gc.gather_inputs(parser)

        params = [
            ("Countries", countries),
            ("Forecast year", str(current_year)),
            ("Lookback years", str(n_years)),
            ("Seasons", f"{outlook_seasons[0]}-{outlook_seasons[-1]}"),
            ("Aggregation", aggregation),
            ("Pooled", str(pool_countries_flag)),
            ("DB", parser.get("DEFAULT", "db")),
            ("Total combinations", str(len(inputs))),
        ]
        ut.display_run_summary("GeoCIF Yield Outlook", params, wait=10)

        if pool_countries_flag:
            gc.execute_models(inputs, logger_obj, parser, loop_fn=gc.loop_execute_pooled)
        else:
            gc.execute_models(inputs, logger_obj, parser)

        # Restore original config values
        for country, orig in originals.items():
            parser.set(country, "forecast_seasons", orig)
        parser.set("DEFAULT", "experiment_name", experiment_name)
        parser.set("DEFAULT", "db", orig_db)
        if orig_parallel is not None:
            parser.set("DEFAULT", "do_parallel_ml", orig_parallel)
        elif parser.has_option("DEFAULT", "do_parallel_ml"):
            parser.remove_option("DEFAULT", "do_parallel_ml")

    # ---- Step 2: Load shapefiles ----
    dg, dict_config = _load_shapefiles(parser)

    # ---- Step 3: Query DB, compute outlook, generate maps ----
    project_name = parser.get("DEFAULT", "project_name", fallback="geocif")
    dir_output = Path(parser.get("PATHS", "dir_output")) / project_name
    if reuse_db is not None:
        db_path = Path(reuse_db)
    else:
        db_path = dir_output / "ml" / "db" / outlook_db

    today = ar.utcnow().to("America/New_York").format("MMMM_DD_YYYY")
    dir_outlook = dir_output / "ml" / "analysis" / today / "outlook"
    os.makedirs(dir_outlook, exist_ok=True)

    all_outlook_frames = []

    for country_crop, config in dict_config.items():
        crop = config["crops"]
        country = country_crop.replace(f"_{crop}", "")
        is_pooled = country == "pooled"
        models = config["models"]

        for model in models:
            logger.info("Yield outlook: %s %s %s", country, crop, model)

            df = _query_predictions(db_path, country_crop, model, experiment_name="outlook")
            if df.empty:
                logger.warning("No predictions found for %s %s %s", country, crop, model)
                continue

            # Get all stages available for current year
            df_current = df[df["Harvest Year"] == current_year]
            if df_current.empty:
                logger.warning(
                    "No predictions for year %d in %s %s %s",
                    current_year, country, crop, model,
                )
                continue

            if is_pooled:
                annotate = False
                map_countries = countries
            else:
                annotate = parser.getboolean(
                    country, "annotate_regions", fallback=False
                )
                map_countries = [country]

            # Compute outlook index
            df_outlook = _compute_outlook_index(
                df, current_year, n_years, aggregation,
                use_latest_stage=use_latest_stage,
            )
            if df_outlook.empty:
                logger.warning(
                    "Could not compute outlook for %s %s %s",
                    country, crop, model,
                )
                continue

            # Use current year's stage name for labeling
            stage_name = df_current["Stage Name"].iloc[-1]

            n_hist = len(
                df[
                    (df["Harvest Year"] < current_year)
                    & (df["Harvest Year"] >= current_year - n_years)
                ]["Harvest Year"].unique()
            )
            if n_hist < 3:
                logger.warning(
                    "Only %d historical years for %s %s %s (requested %d)",
                    n_hist, country, crop, model, n_years,
                )

            df_outlook["Crop"] = crop
            df_outlook["Model"] = model
            df_outlook["Stage Name"] = stage_name
            df_outlook["Forecast Year"] = current_year
            all_outlook_frames.append(df_outlook)

            # Generate map
            _generate_outlook_map(
                dg,
                df_outlook,
                map_countries,
                crop,
                model,
                current_year,
                n_years,
                aggregation,
                dir_outlook,
                stage_name=stage_name,
                annotate_regions=annotate,
            )
            logger.info(
                "Map saved: %s",
                dir_outlook
                / f"yield_outlook_{country}_{crop}_{model}_{stage_name}_{current_year}.png",
            )

    # ---- Consolidated maps (all countries on one map) ----
    if all_outlook_frames:
        df_all = pd.concat(all_outlook_frames, ignore_index=True)
        for (crop, model), df_group in df_all.groupby(
            ["Crop", "Model"]
        ):
            countries_with_data = df_group["Country"].unique().tolist()
            if len(countries_with_data) <= 1:
                continue
            _generate_outlook_map(
                dg,
                df_group,
                countries_with_data,
                crop,
                model,
                current_year,
                n_years,
                aggregation,
                dir_outlook,
                stage_name="combined",
                annotate_regions=False,
            )

    # Save combined CSV
    if all_outlook_frames:
        df_all = pd.concat(all_outlook_frames, ignore_index=True)
        csv_path = dir_outlook / f"yield_outlook_{current_year}.csv"
        df_all.to_csv(csv_path, index=False)
        logger.info("Outlook CSV saved to %s", csv_path)
    else:
        logger.warning("No outlook data generated — check DB has predictions.")


if __name__ == "__main__":
    run()
